import logging

log = logging.getLogger(__name__)

ACCELDATA_LINEAGE_URL = "acceldata_lineage_url"
ACCELDATA_LINEAGE_ENDPOINT = "acceldata_lineage_endpoint"
ACCELDATA_ACCESS_KEY = "acceldata_access_key"
ACCELDATA_SECRET_KEY = "acceldata_secret_key"
ACCELDATA_BEARER_TOKEN = "acceldata_bearer_token"
ACCELDATA_EXPIRES_AT = "acceldata_expires_at"
DATAPROC_CREATE_BATCH_OPERATOR = "DataprocCreateBatchOperator"
RUNTIME_CONFIG = "runtime_config"



def _fetch_open_lineage_spark_properties(parent_run_id, job_name):
    from airflow.models import Variable
    log.info("OPEN_LINEAGE_POLICY Setting up openlineage spark properties")
    acceldata_lineage_url = Variable.get(ACCELDATA_LINEAGE_URL, None)
    acceldata_lineage_endpoint = Variable.get(ACCELDATA_LINEAGE_ENDPOINT, None)
    acceldata_secret_key = Variable.get(ACCELDATA_SECRET_KEY, None)
    acceldata_access_key = Variable.get(ACCELDATA_ACCESS_KEY, None)

    if acceldata_lineage_url is not None and acceldata_lineage_endpoint is not None and acceldata_secret_key is not None and acceldata_access_key is not None:
        properties = {"spark.extraListeners": 'io.openlineage.spark.agent.OpenLineageSparkListener',
                      "spark.openlineage.transport.url": acceldata_lineage_url,
                      "spark.openlineage.parentRunId": parent_run_id,
                      "spark.jars.packages": 'io.openlineage:openlineage-spark:1.1.0',
                      "spark.openlineage.transport.type": 'http',
                      "spark.openlineage.parentJobName": job_name,
                      "spark.openlineage.transport.headers.accessKey": acceldata_access_key,
                      "spark.openlineage.transport.headers.secretKey": acceldata_secret_key,
                      "spark.openlineage.transport.endpoint": acceldata_lineage_endpoint}
        log.info("OPEN_LINEAGE_POLICY Spark Properties: %s ", properties)
        return properties
    else:
        log.error("OPEN_LINEAGE_POLICY Necessary parameters to construct spark properties not available in cache")


# Building batch->runtime_config->properties
def _build_properties(task, parent_run_id, runtime_config):
    spark_properties = _fetch_open_lineage_spark_properties(parent_run_id, task.task_id)
    log.info("OPEN_LINEAGE_POLICY spark_properties: %s", str(spark_properties))
    # If task.batch["runtime_config] contains properties key
    if "properties" in runtime_config:
        existing_properties = runtime_config["properties"]
        # Merge existing and openlineage properties
        merged_properties = {**spark_properties, **existing_properties}
        task.batch["runtime_config"]["properties"] = merged_properties
        log.info("OPEN_LINEAGE_POLICY Properties %s :", merged_properties)
    else:
        #  If task.batch["runtime_config] doesn't have properties key
        task.batch["runtime_config"]["properties"] = spark_properties
        log.info("OPEN_LINEAGE_POLICY Properties %s :", spark_properties)


def _mutate_data_proc_create_batch_operator(task):
    log.info("OPEN_LINEAGE_POLICY mutate_data_proc_create_batch_operator")
    try:
        if hasattr(task, "batch"):
            task_batch = task.batch
            parent_run_id = "{{ macros.OpenLineagePlugin.lineage_run_id(task, task_instance) }}"
            log.debug("OPEN_LINEAGE_POLICY task_batch %s", str(task_batch))
            if RUNTIME_CONFIG in task_batch:
                # Append to run_time config if exists
                runtime_config = task_batch["runtime_config"]
                log.info("OPEN_LINEAGE_POLICY task.batch['runtime_config'] found: %s", str(runtime_config))
                _build_properties(task, parent_run_id, runtime_config)
            else:
                # construct runtime_config here
                task.batch["runtime_config"] = {}
                runtime_config = task.batch["runtime_config"]
                _build_properties(task, parent_run_id, runtime_config)
        log.info("OPEN_LINEAGE_POLICY DataprocCreateBatchOperator mutations completed")
    except Exception as e:
        log.error("OPEN_LINEAGE_POLICY Error when calling task_policy")
        log.error("OPEN_LINEAGE_POLICY %s", e)



from datetime import datetime
from airflow.models.baseoperator import BaseOperator

from acceldata_airflow_sdk.utils.torch_client import TorchDAGClient
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_airflow_sdk.utils.constants import PIPELINE_UID_XCOM, get_dag_run_pipeline_run_id, CONNECTION_ID

import logging

LOGGER = logging.getLogger("airflow.task")


class SpanOperator(BaseOperator):
    """
    Description:
        Used to send span start and end event for any std airflow operator. Just wrap your operator with span operator.
        Make sure you do not add your task in dag(dag parameter should not be specified in the operator being wrapped by
         span operator). If you wrap it using span operator, will take care of that task operator.

    You need to add extra parameter mentioned below. Other parameters will be same as std airflow base operator's parameters

    :param operator: std task operator defined
    :param span_uid: span uid for the task
    :param associated_job_uids: list of job uids
    :param xcom_to_event_mapper_ids: list of xcom keys. Used to send xcom variables in span event job

    """
    def __init__(self, *, operator: BaseOperator, span_uid: str = None, associated_job_uids = None, xcom_to_event_mapper_ids= None, **kwargs):
        """
        You need to add extra parameter mentioned below. Other parameters will be same as std airflow base operator's parameters
        :param operator: std task operator defined
        :param span_uid: span uid for the task
        :param associated_job_uids: list of job uids
        :param xcom_to_event_mapper_ids: list of xcom keys. Used to send xcom variables in span event
        Example :

        --> Defined std operator.

        postgres_operator = PostgresOperator(
            task_id="task_name",
            postgres_conn_id='example_db',
            sql="select * from information_schema.attributes",
        )


        --> To wrap operator with span. Write assign this to your dag (not your std operator)

        span_operator = SpanOperator(
            task_id='task_name',
            span_uid='span.uid',
            operator=postgres_operator,
            dag=dag
        )

        """
        if kwargs.get("provide_context"):
            kwargs.pop('provide_context', None)
        super().__init__(**kwargs)
        self.operator = operator
        self.pipeline_uid = None
        self.span_uid = span_uid
        self.parent_span_ctxt = None
        if associated_job_uids is None:
            self.associated_job_uids = []
        else:
            self.associated_job_uids = associated_job_uids
        self.xcom_to_event_mapper_ids = xcom_to_event_mapper_ids

    def execute(self, context):
        try:
            LOGGER.info("Send span start event")
            task_instance = context['ti']
            parent_span_context = task_instance.xcom_pull(key='parent_span_context')
            conn_id = task_instance.xcom_pull(key=CONNECTION_ID)
            if parent_span_context is None:
                LOGGER.debug('sending new request to catalog to get parent span context')
                self.pipeline_uid = task_instance.xcom_pull(key=PIPELINE_UID_XCOM)
                client = TorchDAGClient(conn_id)
                pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
                self.parent_span_ctxt = client.get_root_span(pipeline_uid=self.pipeline_uid, pipeline_run_id=pipeline_run_id)
                # task_instance.xcom_push(key="parent_span_context", value=self.parent_span_ctxt.__dict__)
            else:
                LOGGER.debug('using xcom to get parent span context to send span event')
                self.parent_span_ctxt = parent_span_context
            if self.span_uid is None:
                opame_six = '{:.6}'.format(type(self.operator).__name__)
                self.span_uid = f'{task_instance.task_id}_{opame_six}.span'
            self.span_context = self.parent_span_ctxt.create_child_span(uid=self.span_uid,
                                                                        context_data={'time': str(datetime.now())},
                                                                        associatedJobUids=self.associated_job_uids
                                                                        )
            context['span_context_parent'] = self.span_context
            xcom_context_data = {}
            if self.xcom_to_event_mapper_ids is None:
                self.xcom_to_event_mapper_ids = []
            else:
                for key in self.xcom_to_event_mapper_ids:
                    value = task_instance.xcom_pull(key=key)
                    if value is not None:
                        xcom_context_data[key] = value
            try:
                self.operator.prepare_for_execution().execute(context)
            except Exception as e1:
                if type(e1) == AttributeError:
                    try:
                        self.operator.execute(context)
                    except Exception as e2:
                        LOGGER.error(e2)
                        raise e2
                else:
                    LOGGER.error(e1)
                    raise e1
        except Exception as e:
            LOGGER.error("Send span end failure event")
            exception = e.__dict__
            LOGGER.error(exception)
            self.span_context.send_event(
                GenericEvent(context_data={'status': 'error', 'error_data': str(e), 'time': str(datetime.now()),
                                           'exception_type': str(type(e).__name__)},
                             event_uid=f'{self.span_uid}.error.event'))
            self.span_context.failed(
                context_data=xcom_context_data)
            raise e
        else:
            LOGGER.info("Send span end success event")
            self.span_context.end(context_data=xcom_context_data)

    def set_downstream(self, task_or_task_list) -> None:
        super().set_downstream(task_or_task_list)

    def set_upstream(self, task_or_task_list) -> None:
        super().set_upstream(task_or_task_list)



from __future__ import annotations

from airflow.plugins_manager import AirflowPlugin
from listener import events_listener

import os

# os.environ["TORCH_CATALOG_URL"] = "URL of the torch catalog"
# os.environ["TORCH_ACCESS_KEY"] = "API access key generated from torch UI"
# os.environ["TORCH_SECRET_KEY"] = "API secret key generated from torch UI"

# If set matching dag ids will be ignored and everything else will be observed
# IGNORE and OBSERVE environment variables are mutually exclusive
# Don't set if OBSERVE environment variables are set in the below step
# os.environ["DAGIDS_TO_IGNORE"] = "Comma separated dag ids to ignore observation"
# os.environ["DAGIDS_REGEX_TO_IGNORE"] = "Regex for dag ids to ignore observation"

# If set matching dag ids will be observed and everything else will be ignored.
# IGNORE and OBSERVE environment variables are mutually exclusive
# Don't set if IGNORE environment variables are set in the above step
# os.environ["DAGIDS_TO_OBSERVE"] = "Comma separated dag ids to observe"
# os.environ["DAGIDS_REGEX_TO_OBSERVE"] = "Regex for dag ids to observe"


class AcceldataListenerPlugin(AirflowPlugin):
    name = "AcceldataListenerPlugin"
    listeners = [events_listener]



import os
from airflow.hooks.base import BaseHook
import json
from distutils import util


class Credentials:
    def __init__(self, conn_id):
        connection = BaseHook.get_connection(conn_id)

        self.url = connection.host
        self.access_key = connection.login
        self.secret_key = connection.password
        if connection.extra is not None and len(connection.extra) > 0:
            version_check = json.loads(connection.extra)['ENABLE_VERSION_CHECK']
        else:
            version_check = True
        if isinstance(version_check, str):
            self.do_version_check = bool(util.strtobool(version_check))
        else:
            self.do_version_check = version_check


# setup these 4 env vars in your airflow environment. You can create api keys from torch ui's setting page.
def torch_credentials(conn_id=None):
    if conn_id is None:
        creds = {
            'url': os.getenv('TORCH_CATALOG_URL', 'https://torch.acceldata.local:5443'),
            'access_key': os.getenv('TORCH_ACCESS_KEY', 'OY2VVIN2N6LJ'),
            'secret_key': os.getenv('TORCH_SECRET_KEY', 'da6bDBimQfXSMsyyhlPVJJfk7Zc2gs'),
            'do_version_check': os.getenv('ENABLE_VERSION_CHECK', True)
        }
    else:
        creds = Credentials(conn_id).__dict__
    return creds



from __future__ import annotations
from airflow.listeners import hookimpl
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.utils.state import TaskInstanceState

import logging
from listener.utils import create_pipeline_for_dag_run, \
    get_torch_client, end_pipeline_for_dag_run, create_pipeline_for_older_versions, create_job_and_span, \
    end_job_and_span, end_pipeline_for_older_versions, observe_dag_env_variables_detected, \
    ignore_dag_env_variables_detected, should_observe_dag, should_ignore_dag, validate_and_auto_instrument_dag
from listener.constants import INVALID_ENV_SELECTION_MESSAGE

from acceldata_sdk.models.pipeline import PipelineRunResult, PipelineRunStatus

log = logging.getLogger(__name__)


@hookimpl
def on_dag_run_running(dag_run: DagRun, msg: str):
    log.info("Listener Event==> on_dag_run_running")
    try:
        dag_id = dag_run.dag_id
        if observe_dag_env_variables_detected() and ignore_dag_env_variables_detected():
            log.error(INVALID_ENV_SELECTION_MESSAGE)
            return

        if validate_and_auto_instrument_dag(dag_id):
            log.info("Observing the DAG with dag_id  %s ", dag_id)
            create_pipeline_for_dag_run(dag_run, get_torch_client(), msg)
        else:
            log.info("Instrumentation not enabled for the dag with dag_id %s ", dag_id)
    except Exception as e:
        log.error("Error in on_dag_run_running: %s", e)


@hookimpl
def on_dag_run_failed(dag_run: DagRun, msg: str):
    log.info("Listener Event==> on_dag_run_failed")
    try:
        dag_id = dag_run.dag_id
        if observe_dag_env_variables_detected() and ignore_dag_env_variables_detected():
            log.error(INVALID_ENV_SELECTION_MESSAGE)
            return

        if validate_and_auto_instrument_dag(dag_id):
            log.info("Observing the DAG with dag_id  %s ", dag_id)
            end_pipeline_for_dag_run(dag_run, get_torch_client(), PipelineRunResult.FAILURE, PipelineRunStatus.FAILED,
                                     msg)
        else:
            log.info("Instrumentation not enabled for the dag with dag_id %s ", dag_id)
    except Exception as e:
        log.error("Error in on_dag_run_failed: %s", e)


@hookimpl
def on_dag_run_success(dag_run: DagRun, msg: str):
    log.info("Listener Event==> on_dag_run_success")
    try:
        dag_id = dag_run.dag_id
        if observe_dag_env_variables_detected() and ignore_dag_env_variables_detected():
            log.error(INVALID_ENV_SELECTION_MESSAGE)
            return

        if validate_and_auto_instrument_dag(dag_id):
            log.info("Observing the DAG with dag_id  %s ", dag_id)
            end_pipeline_for_dag_run(dag_run, get_torch_client(), PipelineRunResult.SUCCESS,
                                     PipelineRunStatus.COMPLETED, msg)
        else:
            log.info("Instrumentation not enabled for the dag with dag_id %s ", dag_id)
    except Exception as e:
        log.error("Error in on_dag_run_success: %s", e)


@hookimpl
def on_task_instance_running(previous_state: TaskInstanceState, task_instance: TaskInstance, session):
    log.info("Listener Event==> on_task_instance_running")
    try:
        dag_id = task_instance.dag_id
        if observe_dag_env_variables_detected() and ignore_dag_env_variables_detected():
            log.error(INVALID_ENV_SELECTION_MESSAGE)
            return

        if validate_and_auto_instrument_dag(dag_id):
            log.info("Observing the DAG with dag_id  %s ", dag_id)
            create_job_and_span(task_instance, get_torch_client())
        else:
            log.info("Instrumentation not enabled for the dag with dag_id %s ", dag_id)
    except Exception as e:
        log.error("Error in on_task_instance_running: %s", e)


@hookimpl
def on_task_instance_success(previous_state: TaskInstanceState, task_instance: TaskInstance, session):
    log.info("Listener Event==> on_task_instance_success")
    try:
        dag_id = task_instance.dag_id
        if observe_dag_env_variables_detected() and ignore_dag_env_variables_detected():
            log.error(INVALID_ENV_SELECTION_MESSAGE)
            return

        if validate_and_auto_instrument_dag(dag_id):
            log.info("Observing the DAG with dag_id  %s ", dag_id)
            end_job_and_span(task_instance, get_torch_client())
            log.info("Pushed span to Torch Server for success task")
            # Versions prior to 2.5.0 should end pipeline for task with zero downstream events
            end_pipeline_for_older_versions(task_instance, PipelineRunResult.SUCCESS, PipelineRunStatus.COMPLETED,
                                            session)
        else:
            log.info("Instrumentation not enabled for the dag with dag_id %s ", dag_id)
    except Exception as e:
        log.error("Error in on_task_instance_success: %s", e)


@hookimpl
def on_task_instance_failed(previous_state: TaskInstanceState, task_instance: TaskInstance, session):
    log.info("Listener Event==> on_task_instance_failed")
    try:
        dag_id = task_instance.dag_id
        if observe_dag_env_variables_detected() and ignore_dag_env_variables_detected():
            log.error(INVALID_ENV_SELECTION_MESSAGE)
            return

        if validate_and_auto_instrument_dag(dag_id):
            log.info("Observing the DAG with dag_id  %s ", dag_id)
            end_job_and_span(task_instance, get_torch_client())
            log.info("Pushed span to Torch Server for failed task")
            # Versions prior to 2.5.0 should end pipeline for events with zero downstream events or upon failure
            end_pipeline_for_older_versions(task_instance, PipelineRunResult.FAILURE, PipelineRunStatus.FAILED, session)
        else:
            log.info("Instrumentation not enabled for the dag with dag_id %s ", dag_id)
    except Exception as e:
        log.error("Error in on_task_instance_failed: %s", e)



PIPELINE_RUN_ID = "pipeline_run_id"
root_span_suffix = "_root_span"
STARTED = "Started"
ENDED = "Ended"
DURATION = "Duration"
TRIGGER_RULE = "Trigger Rule"
PIPELINE_NAME_SUFFIX = "_pipeline"
PIPELINE_STATE = "pipeline_state"
ROOT_SPAN_SUFFIX = "_root_span"
STATE = "state"
FAILED = "failed"
SUCCESS = "success"
MESSAGE = "message"
DAG_ID = "airflow_dag_id"
TASK_ID = "airflow_task_id"
DAG_RUN_ID = "airflow_dag_run_id"
TASK_RUN_ID = "airflow_task_run_id"
OPERATOR = "operator"
TRY_NUMBER = "try_number"
MAP_INDEX = "map_index"
DURATION_UNIT = " seconds"
RUN_TYPE = "run_type"
SCHEDULE_INTERVAL = "schedule_interval"
OWNER = "owner"
QUEUE = "queue"
DAG_URL = "url"
INVALID_ENV_SELECTION_MESSAGE = "Environment variables --OBSERVE DAG and --IGNORE DAG are mutually exclusive.  " \
                                " Please remove one option to observe/ignore the DAG. "



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import traceback
from typing import List, Optional

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.airflow.utils import get_job_name, try_import_from_string
from openlineage.client.facet import SqlJobFacet
from openlineage.common.provider.bigquery import BigQueryDatasetsProvider, BigQueryErrorRunFacet
import logging

_BIGQUERY_NAMESPACE = 'bigquery'

log = logging.getLogger(__name__)

_BIGQUERY_CONN_URL = 'bigquery'


class BigQueryInsertJobExtractor(BaseExtractor):
    def __init__(self, operator):
        super().__init__(operator)

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ['BigQueryInsertJobOperator']

    def extract(self) -> Optional[TaskMetadata]:
        return None

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        self.log.debug(f"BigQueryInsertJobOperator extract_on_complete({task_instance})")
        try:
            bigquery_job_id = task_instance.xcom_pull(task_ids=task_instance.task_id, key=None)
            if bigquery_job_id is None:
                raise Exception(
                    "Xcom could not resolve BigQuery job id. Job may have failed."
                )
        except Exception as e:
            self.log.error(
                f"Cannot retrieve job details from BigQuery.Client. {e}", exc_info=True
            )
            return TaskMetadata(
                name=get_job_name(task=self.operator),
                run_facets={
                    "bigQuery_error": BigQueryErrorRunFacet(
                        clientError=f"{e}: {traceback.format_exc()}",
                    )
                }
            )

        client = self._get_client()

        stats = BigQueryDatasetsProvider(client=client).get_facets(bigquery_job_id)
        inputs = stats.inputs
        output = stats.output

        for ds in inputs:
            ds.input_facets = self._get_input_facets()

        run_facets = stats.run_facets
        self.operator.sql = self.operator.configuration['query']['query']
        self.operator.use_legacy_sql = self.operator.configuration['query']['useLegacySql']
        log.debug("SQL Query : %s", str(self.operator.sql))
        job_facets = {
            "sql": SqlJobFacet(self.operator.sql)
        }

        return TaskMetadata(
            name=get_job_name(task=self.operator),
            inputs=[ds.to_openlineage_dataset() for ds in inputs],
            outputs=[output.to_openlineage_dataset()] if output else [],
            run_facets=run_facets,
            job_facets=job_facets
        )

    def _get_client(self):
        # lazy-load the bigquery Client due to its slow import
        from google.cloud.bigquery import Client
        # Get client using Airflow hook - this way we use the same credentials as Airflow
        if hasattr(self.operator, 'hook') and self.operator.hook:
            hook = self.operator.hook
            return hook.get_client(
                project_id=hook.project_id,
                location=hook.location
            )
        BigQueryHook = try_import_from_string(
            'airflow.providers.google.cloud.operators.bigquery.BigQueryHook'
        )
        if BigQueryHook is not None:
            params = {
                "gcp_conn_id": self.operator.gcp_conn_id,
                # "use_legacy_sql": self.operator.use_legacy_sql,
                "location": self.operator.location,
                "impersonation_chain": self.operator.impersonation_chain,
            }
            if hasattr(self.operator, "delegate_to"):
                params["delegate_to"] = self.operator.delegate_to

            hook = BigQueryHook(**params)
            return hook.get_client(
                project_id=hook.project_id,
                location=hook.location
            )
        return Client()

    def _get_input_facets(self):
        return {}



from acceldata_airflow_sdk.utils.torch_client import TorchDAGClient
from acceldata_sdk.models.job import CreateJob, JobMetadata
from datetime import datetime

import logging

LOGGER = logging.getLogger("airflow.task")
PIPELINE_UID_XCOM = 'pipeline_uid_ff069534-5069-45b1-b737-aea6229db4bf'
CONNECTION_ID = 'CONNECTION_ID'


def get_dag_run_pipeline_run_id(task_instance):
    return task_instance.xcom_pull(key=f'{task_instance.dag_id}_pipeline_run_id')


def create_job_span(task_instance, job_uid, inputs, outputs, metadata, context_job, bounded_by_span,
                    xcom_to_event_mapper_ids, span_uid, kwargs):
    try:
        xcom_context_data = {}
        span_uid_temp = span_uid
        conn_id = task_instance.xcom_pull(key=CONNECTION_ID)
        pipeline_uid_ = task_instance.xcom_pull(key=PIPELINE_UID_XCOM)
        client = TorchDAGClient(conn_id)
        pipeline = client.get_pipeline(pipeline_uid_)
        pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
        pipeline_run = pipeline.get_run(pipeline_run_id)
        LOGGER.info("Creating job")
        job = CreateJob(
            uid=job_uid,
            name=f'{job_uid} Job',
            pipeline_run_id=pipeline_run.id,
            description=f'{job_uid} created using torch job decorator',
            inputs=inputs,
            outputs=outputs,
            meta=metadata,
            context=context_job
        )
        job = pipeline.create_job(job)

    except Exception as e:
        LOGGER.error("Error in creating job")
        exception = e.__dict__
        LOGGER.error(exception)
        raise e
    else:
        LOGGER.info("Successfully created job.")
        if bounded_by_span:
            LOGGER.info('Sending Span Start event')
            try:
                xcoms = xcom_to_event_mapper_ids
                parent_span_context = task_instance.xcom_pull(key='parent_span_context')
                if parent_span_context is None:
                    LOGGER.debug('Sending new request to catalog to get parent span context')
                    parent_span_context = client.get_root_span(pipeline_uid=pipeline_uid_)
                else:
                    LOGGER.debug('Using xcom to get parent span context to send span event')
                associated_job_uids = [job_uid]
                if span_uid is None:
                    span_uid_temp = job_uid
                span_context = parent_span_context.create_child_span(
                    uid=span_uid_temp,
                    context_data={
                        'time': str(datetime.now()),
                        'xcom_to_event_mapper_ids': xcoms
                    },
                    associatedJobUids=associated_job_uids)
                if xcoms is None:
                    xcoms = []
                else:
                    for key in xcoms:
                        value = task_instance.xcom_pull(key=key)
                        if value is not None:
                            xcom_context_data[key] = value
                    LOGGER.info('Xcom context data ', xcom_context_data)
                kwargs['span_context_parent'] = span_context

            except Exception as ex:
                print(f'Span creation failed with exception{str(ex)}')
    return span_uid_temp, kwargs, xcom_context_data


def is_pipeline_run_ended(pipeline_run):
    if pipeline_run.status == "COMPLETED" or pipeline_run.status == "FAILED" or pipeline_run.status == "ABORTED":
        return True



from __future__ import annotations
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.models import DagBag, DagModel
from airflow.configuration import conf
from pkg_resources import parse_version

import logging
import os
import functools
import airflow
import re

from acceldata_sdk.torch_client import TorchClient
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.models.pipeline import CreatePipeline, PipelineMetadata, PipelineRunResult, PipelineRunStatus, \
    PipelineRun
from acceldata_sdk.errors import APIError, TorchSdkException
from acceldata_sdk.events.generic_event import GenericEvent

from listener.constants import PIPELINE_NAME_SUFFIX, ROOT_SPAN_SUFFIX, STARTED, PIPELINE_STATE, ENDED, \
    STATE, \
    DURATION, SUCCESS, FAILED, MESSAGE, DAG_ID, DAG_RUN_ID, TASK_RUN_ID, OPERATOR, TRY_NUMBER, MAP_INDEX, TASK_ID, \
    DURATION_UNIT, RUN_TYPE, OWNER, QUEUE, SCHEDULE_INTERVAL, DAG_URL

log = logging.getLogger(__name__)


def singleton(cls):
    @functools.wraps(cls)
    def wrapper(*args, **kwargs):
        if not wrapper.instance:
            log.info("Initializing singleton instance ")
            wrapper.instance = cls(*args, **kwargs)
        return wrapper.instance

    wrapper.instance = None
    return wrapper


@singleton
class SingletonTorchClient:
    def __init__(self) -> None:
        self.torchClient = get_or_create_torch_client()

    def get_torch_client(self):
        return self.torchClient


def get_dag_run_url(dag_run):
    base_url=conf.get('webserver', 'BASE_URL')
    import urllib.parse
    execution_date = urllib.parse.quote(dag_run.execution_date.isoformat())
    dag_url = f'{base_url}/graph?root=&dag_id={dag_run.dag_id}&execution_date={execution_date}&arrang=LR'
    log.info(dag_url)
    return dag_url


def get_or_create_torch_client():
    #TODO: Required on Mac else the app crashes
    os.environ["no_proxy"] = "*"
    torch_catalog_url = os.environ["TORCH_CATALOG_URL"]
    access_key = os.environ["TORCH_ACCESS_KEY"]
    secret_key = os.environ["TORCH_SECRET_KEY"]
    torch_client = TorchClient(url=torch_catalog_url, access_key=access_key,
                               secret_key=secret_key)
    return torch_client


def should_ignore_dag(dag_id):
    dagids_to_ignore = os.environ.get("DAGIDS_TO_IGNORE")
    dagids_regex_to_ignore = os.environ.get("DAGIDS_REGEX_TO_IGNORE")
    ignore_dag_id = False

    # If any one of these env variables are present, ignore based on the criteria
    dag_id_match = False
    dag_id_regex_match = False
    if dagids_regex_to_ignore is not None:
        dag_id_regex_match = re.search(dagids_regex_to_ignore, dag_id)

    if dagids_to_ignore is not None:
        if dag_id in dagids_to_ignore.split(','):
            dag_id_match = True

    if dag_id_regex_match or dag_id_match:
        ignore_dag_id = True
        log.info("Ignoring dag_id %s instrumentation", dag_id)

    return ignore_dag_id


def observe_dag_env_variables_detected():
    dagids_to_observe = os.environ.get("DAGIDS_TO_OBSERVE")
    dagids_regex_to_observe = os.environ.get("DAGIDS_REGEX_TO_OBSERVE")
    log.debug("DAGIDS_TO_OBSERVE")
    log.debug(dagids_to_observe)
    log.debug("DAGIDS_REGEX_TO_OBSERVE")
    log.debug(dagids_regex_to_observe)
    if dagids_to_observe is not None or dagids_regex_to_observe is not None:
        return True
    return False


def ignore_dag_env_variables_detected():
    dagids_to_ignore = os.environ.get("DAGIDS_TO_IGNORE")
    dagids_regex_to_ignore = os.environ.get("DAGIDS_REGEX_TO_IGNORE")
    if dagids_to_ignore is not None or dagids_regex_to_ignore is not None:
        return True
    return False


def should_observe_dag(dag_id):
    dagids_to_observe = os.environ.get("DAGIDS_TO_OBSERVE")
    dagids_regex_to_observe = os.environ.get("DAGIDS_REGEX_TO_OBSERVE")

    # If any one of these env variables are present, observe based on the criteria
    observe_dag_id = False
    dag_id_match = False
    dag_id_regex_match = False

    if dagids_regex_to_observe is not None:
        dag_id_regex_match = re.search(dagids_regex_to_observe, dag_id)

    if dagids_to_observe is not None:
        if dag_id in dagids_to_observe.split(','):
            dag_id_match = True

    if dag_id_regex_match or dag_id_match:
        observe_dag_id = True
        log.info("Observing dag_id %s ", dag_id)

    return observe_dag_id


def get_torch_client() -> TorchClient:
    singleton_torch_client = SingletonTorchClient()
    return singleton_torch_client.get_torch_client()


def is_dag_run_supported_for_version(version):
    parsed_version_airflow = parse_version(version)
    parsed_version_dag_run_support = parse_version("2.5.0.dev0")
    log.debug(f"parsed_version_airflow : {parsed_version_airflow}")
    log.debug(f"parsed_version_dag_run_support : {parsed_version_dag_run_support}")
    if parsed_version_airflow >= parsed_version_dag_run_support:
        log.debug("dag_run state change events are supported ")
        return True
    log.info("dag_run state change events not supported ")
    return False


def get_downstream_ids_task_instance(task_instance: TaskInstance, session):
    dag_model = DagModel.get_current(task_instance.dag_id, session=session)
    dagbag = DagBag(dag_folder=dag_model.fileloc, read_dags_from_db=True)
    dag = dagbag.get_dag(task_instance.dag_id, session=session)
    task_ids = dag.task_dict[task_instance.task_id].downstream_task_ids
    return task_ids


def create_pipeline_for_dag_run(dag_run: DagRun, torch_client, msg: str):
    log.debug("create_pipeline_for_dag_run invoked")
    dag_id = dag_run.dag_id
    pipeline_name = dag_id
    root_span_uid = dag_id + ROOT_SPAN_SUFFIX
    dag_url = get_dag_run_url(dag_run)
    try:
        # 1. Create Pipeline
        log.info("Creating pipeline for the DAG with the pipeline_name %s", pipeline_name)
        pipeline = CreatePipeline(uid=pipeline_name, name=pipeline_name)
        pipeline_response = torch_client.create_pipeline(pipeline)
        dag_run_id = dag_run.run_id

        # 2. Create pipeline run
        log.info("Creating pipeline run for the pipeline with name %s", pipeline_name)
        log.debug(f"dag run {dag_run}")
        context_data = {
            STARTED: str(dag_run.start_date),
            STATE: str(dag_run.state),
            DAG_ID: str(dag_id),
            DAG_RUN_ID: str(dag_run_id),
            RUN_TYPE: str(dag_run.run_type),
            DAG_URL: dag_url,
            # TODO: Code breaks for the older version,
            # SCHEDULE_INTERVAL: str(dag_run.dag.schedule_interval),
            MESSAGE: msg}

        log.info("Getting the pipeline run corresponding to dag run based on continuation_id and pipeline_id")
        continuation_id = f"{dag_id}.{dag_run_id}"
        pipeline_run = pipeline_response.create_pipeline_run(context_data=context_data, continuation_id=continuation_id)
        pipeline_run_id = pipeline_run.id
        log.info("Created pipeline run with the pipeline_run_id: %s", pipeline_run_id)

        # 3. Create root span
        log.info("Creating the root span")
        root_span = pipeline_run.create_span(uid=root_span_uid)
        event_uid = f'{root_span_uid}.start_event'
        pipeline_start_event = GenericEvent(context_data=context_data,
                                            event_uid=event_uid)
        root_span.send_event(span_event=pipeline_start_event)
    except APIError as api_error:
        # 2. Pipeline doesn't exist , create pipeline with the dag id as the identity
        log.error("Pipeline creation for the pipeline : %s  failed with error:  %s", str(pipeline_name), str(api_error))


def end_pipeline_for_dag_run(dag_run: DagRun, torch_client, pipeline_run_result: PipelineRunResult,
                             pipeline_run_status: PipelineRunStatus, msg: str):
    dag_id = dag_run.dag_id
    pipeline_name = dag_id
    dag_run_id = dag_run.run_id
    log.debug("end_pipeline_for_dag_run called")
    # 1. get the pipeline corresponding to the above pipelineName
    log.debug("Getting the pipeline corresponding to the pipeline name: %s", pipeline_name)
    pipeline = torch_client.get_pipeline(pipeline_name)

    #  2. get the pipeline run for the dag run
    log.info("Getting the pipeline run corresponding to dag run based on continuation_id and pipeline_id")
    continuation_id = f"{dag_id}.{dag_run_id}"
    pipeline_run = torch_client.get_pipeline_run(continuation_id=continuation_id,
                                                 pipeline_id=pipeline.id)

    # 3. get the root span for the run and mark it as success or failure accordingly
    log.debug("Getting the root span for the pipeline run with pipeline run id: %s", pipeline_run.id)
    root_span_context = pipeline_run.get_root_span()
    log.debug("Got the root span context %s", root_span_context)

    root_span_uid = dag_run.dag_id + ROOT_SPAN_SUFFIX
    log.debug("Root span uid: %s", root_span_uid)
    event_uid = f'{root_span_uid}.end_event'
    context_data = {STARTED: str(dag_run.start_date),
                    ENDED: str(dag_run.end_date),
                    DAG_ID: str(dag_run.dag_id),
                    DAG_RUN_ID: str(dag_run.run_id),
                    RUN_TYPE: str(dag_run.run_type),
                    STATE: str(dag_run.state),
                    MESSAGE: msg}
    pipeline_end_event = GenericEvent(context_data=context_data,
                                      event_uid=event_uid)

    # 4. Mark the run as success or failure accordingly
    if pipeline_run_result == PipelineRunResult.SUCCESS:
        log.debug("pipeline_run_result is: %s", pipeline_run_result)
        root_span_context.end()
        log.info("Root span has been ended.")

    elif pipeline_run_result == PipelineRunResult.FAILURE:
        log.debug("pipeline_run_result is: %s", pipeline_run_result)
        root_span_context.failed()
        log.info("Root span has been Failed.")

    log.info("Ending the pipeline run")
    root_span_context.send_event(span_event=pipeline_end_event)
    pipeline_run.update_pipeline_run(context_data=context_data,
                                     result=pipeline_run_result,
                                     status=pipeline_run_status)


def validate_and_auto_instrument_dag(dag_id):
    # Observe the given dag if any one of below condition happens:
    observe_dag_var_detected = observe_dag_env_variables_detected()
    ignore_dag_var_detected = ignore_dag_env_variables_detected()
    # 1. None of the environment variables to observe/ignore are present
    # 2. dag_id matches the OBSERVE environment variable criteria
    # 3. dag_id doesn't match the IGNORE environment variable criteria

    if ((not observe_dag_var_detected and not ignore_dag_var_detected)
            or (observe_dag_var_detected and should_observe_dag(dag_id))
            or (ignore_dag_var_detected and not should_ignore_dag(dag_id))):
        return True
    return False


def create_job_and_span(task_instance: TaskInstance, torch_client: TorchClient):
    dag_run_id = task_instance.dag_run.run_id
    dag_id = task_instance.dag_id
    pipeline_name = str(task_instance.dag_id)
    pipeline = torch_client.get_pipeline(pipeline_name)
    log.info("Getting the pipeline run corresponding to dag run based on continuation_id and pipeline_id")
    continuation_id = f"{dag_id}.{dag_run_id}"
    pipeline_run = torch_client.get_pipeline_run(continuation_id=continuation_id,
                                                 pipeline_id=pipeline.id)
    root_span_context = pipeline_run.get_root_span()
    log.debug("Got the root span context %s", root_span_context)

    # 1. create a job with  job_uid and inputs based on upstream tasks
    upstream_list = task_instance.task.upstream_list
    input_nodes = []
    associated_job_uids = []

    # Construct node inputs
    log.debug("Constructing the input_nodes from upstream tasks")
    for upstream in upstream_list:
        job_uid = upstream.task_id
        input_nodes.append(Node(job_uid=job_uid))

    job_uid = task_instance.task_id
    associated_job_uids.append(job_uid)
    job = CreateJob(
        uid=job_uid,
        name=job_uid,
        version=pipeline_run.versionId,
        pipeline_run_id=pipeline_run.id,
        inputs=input_nodes,
        bounded_by_span=False
    )

    log.info("Creating job %s", job)
    pipeline.create_job(job)

    # 2. get the root span and then create child span on that object using the spanUid above
    # and pass the jobUid as well to this function
    log.info("Initiate Child span for job with job_uid %s", job_uid)
    child_span_uid = f'{job_uid}.span'
    log.info("associatedJobUids %s for the child span with child_span_uid %s", associated_job_uids, child_span_uid)
    child_span = root_span_context.create_child_span(uid=child_span_uid, associatedJobUids=associated_job_uids)
    log.info("Created span with child_span_uid %s", child_span_uid)

    child_span.start()
    log.info("Started the child span with child_span_uid %s", child_span_uid)
    log.info("Sending job start event the child span with child_span_uid %s", child_span_uid)

    event_uid = f'{child_span_uid}.start_event'
    context_data = {TASK_ID: str(task_instance.task_id),
                    TASK_RUN_ID: str(task_instance.run_id),
                    STARTED: str(task_instance.start_date),
                    OPERATOR: str(task_instance.operator),
                    STATE: str(task_instance.state),
                    TRY_NUMBER: str(task_instance.try_number),
                    OWNER: str(task_instance.task.owner),
                    QUEUE: str(task_instance.task.queue)}
    job_start_event = GenericEvent(context_data=context_data,
                                   event_uid=event_uid)

    child_span.send_event(span_event=job_start_event)
    log.info("Successfully sent job start event for the child span with child_span_uid %s", child_span_uid)


def end_job_and_span(task_instance: TaskInstance, torch_client: TorchClient, ):
    log.info("TaskInstance State: %s", task_instance.state)
    # 1. Get pipeline for the task
    log.info("Get pipeline for the task")
    dag_id = task_instance.dag_id
    pipeline_name = str(dag_id)
    pipeline = torch_client.get_pipeline(pipeline_name)

    #  2. get the pipeline run on the dag run
    log.info("Getting the pipeline run for the corresponding dag run based on continuation_id and pipeline_id")
    dag_run_id = task_instance.dag_run.run_id
    continuation_id = f"{dag_id}.{dag_run_id}"
    pipeline_run = torch_client.get_pipeline_run(continuation_id=continuation_id,
                                                 pipeline_id=pipeline.id)
    job_uid = task_instance.task_id

    child_span_uid = f'{job_uid}.span'
    log.info("Fetching the child span with child span uid %s", child_span_uid)
    child_span = pipeline_run.get_span(child_span_uid)

    log.info("Triggering job end events based on the task state")
    context_data = {TASK_ID: str(task_instance.task_id),
                    TASK_RUN_ID: str(task_instance.run_id),
                    STARTED: str(task_instance.start_date),
                    ENDED: str(task_instance.start_date),
                    DURATION: str(task_instance.duration) + DURATION_UNIT,
                    OPERATOR: str(task_instance.operator),
                    STATE: str(task_instance.state),
                    TRY_NUMBER: str(task_instance.try_number)}

    if str(task_instance.state) == FAILED:
        # TODO : try to extract failure reason and logs
        log.info("Task instance in failed State. Sending JobFailEvent")
        event_uid = f'{child_span_uid}.failed_event'
        job_fail_event = GenericEvent(context_data=context_data, event_uid=event_uid)
        child_span.send_event(span_event=job_fail_event)
        child_span.failed()
        log.info("Marking child span as failed")

    if str(task_instance.state) == SUCCESS:
        log.info("Task instance in success State. Sending JobEndEvent")
        event_uid = f'{child_span_uid}.success_event'
        job_end_event = GenericEvent(context_data=context_data, event_uid=event_uid)
        child_span.send_event(span_event=job_end_event)
        child_span.end()
        log.info("Marking child span as success")


def create_pipeline_for_older_versions(task_instance: TaskInstance, session):
    log.info("Checking if pipeline needs to be created for the task_instance : %s", task_instance)
    upstream_size = len(task_instance.task.upstream_list)
    log.debug("Upstream size: %s", upstream_size)
    current_version = airflow.__version__
    log.debug(f"Current Airflow Version : {current_version}")
    if upstream_size == 0:
        if not is_dag_run_supported_for_version(current_version):
            log.info("Creating the pipeline for events with zero upstream events for the Current Airflow Version: %s",
                     current_version)
            create_pipeline_for_dag_run(task_instance.dag_run, get_torch_client(), msg="", )
        else:
            log.debug("task instance initiation handling not needed to start the pipeline for the newer versions")


def end_pipeline_for_older_versions(task_instance: TaskInstance, pipelineRunResult: PipelineRunResult,
                                    pipelineRunStatus: PipelineRunStatus, session):
    log.info("Checking if pipeline needs to be ended for the task instance : %s ", task_instance)
    current_version = airflow.__version__
    log.debug(f"Current Airflow Version : {current_version}")
    task_ids = get_downstream_ids_task_instance(task_instance, session)
    log.debug("downstream task_ids: %s", task_ids)
    downstream_size = len(task_ids)
    log.debug("Downstream length: %s", downstream_size)
    if downstream_size == 0 or task_instance.state == FAILED:
        if not is_dag_run_supported_for_version(current_version):
            log.info("Either the Downstream size is 0 or task_instance is in failed state.")
            log.info("Ending the pipeline run the with the Current Airflow Version: %s",
                     current_version)
            end_pipeline_for_dag_run(task_instance.dag_run, get_torch_client(), pipelineRunResult,
                                     pipelineRunStatus, msg="")
        else:
            log.debug("task instance completion handling not needed to terminate the pipeline for the newer versions")
    else:
        log.debug("Downstream size is > 0 and task_instance state is not FAILED. Proceeding with the next tasks")



from airflow.models import BaseOperator

import logging

log = logging.getLogger(__name__)

from policy.config_helper import _mutate_data_proc_create_batch_operator


# Mutating task level policy
def task_policy(task: BaseOperator):
    try:
        log.info("OPEN_LINEAGE_POLICY task_policy invoked")
        task_obj_type = type(task)
        task_obj_class_name = task_obj_type.__name__
        log.info("OPEN_LINEAGE_POLICY task object class name %s, proceeding with further checks", task_obj_class_name)
        from airflow.providers.google.cloud.operators.dataproc import DataprocCreateBatchOperator

        # apply policy only for DataprocCreateBatchOperator
        if isinstance(task, DataprocCreateBatchOperator):
            log.info("OPEN_LINEAGE_POLICY task is instance of DataprocCreateBatchOperator. Proceeding with mutations")
            _mutate_data_proc_create_batch_operator(task)

    except Exception as e:
        log.error("OPEN_LINEAGE_POLICY Error when calling task_policy")
        log.error("OPEN_LINEAGE_POLICY %s", e)



from setuptools import setup, find_packages

classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Developers',
  'Operating System :: OS Independent',
  'License :: OSI Approved :: Apache Software License',
  'Programming Language :: Python :: 3'
]

setup(
  name='acceldata_airflow_sdk',
  version='3.1.0',
  description='Acceldata Airflow SDK.' + '\n\n' + open('README.txt').read(),
  long_description=open('README.md').read() + '\n\n' + open('CHANGELOG.txt').read(),
  long_description_content_type="text/markdown",
  url='',
  author='acceldata',
  author_email='apisupport@acceldata.io',
  license='MIT License',
  classifiers=classifiers,
  keywords='acceldata-airflow',
  packages=find_packages(),
  install_requires=['requests', 'dataclasses;python_version<"3.7"', 'typing;python_version<"3.5"']
)



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import os
from datetime import datetime, timedelta
from typing import Any
from dateutil import parser
from collections import defaultdict

import requests
import yaml
import os

from openlineage.client.transport.http import TokenProvider
from airflow.models import Variable

log = logging.getLogger(__name__)

# Set the environment variable to bypass proxy
os.environ["no_proxy"] = "*"
CLIENT_ID = "acceldata-app"
TOKEN_PROVIDER_PATH = "/admin/api/onboarding/token-exchange?grant_type=api_keys"
ACCELDATA_LINEAGE_URL = "acceldata_lineage_url"
ACCELDATA_LINEAGE_ENDPOINT = "acceldata_lineage_endpoint"
ACCELDATA_ACCESS_KEY = "acceldata_access_key"
ACCELDATA_SECRET_KEY = "acceldata_secret_key"
ACCELDATA_BEARER_TOKEN = "acceldata_bearer_token"
ACCELDATA_EXPIRES_AT = "acceldata_expires_at"


class AccessKeySecretKeyTokenProvider(TokenProvider):

    def __init__(self, config: dict[str, str]) -> None:
        super().__init__(config)
        log.info("AccessKeySecretKeyTokenProvider constructor invoked. Setting up the access_key and secret_Key")
        self.access_key = config["access_key"]
        self.secret_key = config["secret_key"]

        # Persisting configuration from yaml
        log.info("Inside load config for the task policy")
        if self.is_config_loaded():
            log.info("Config is already loaded. Skipping loading from the config.")
        else:
            log.info("Loading the config")
            config = self.load_config()
            self.persist_config_to_airflow_variables(config)

    # Checking if the acceldata_lineage_backend_url is already available
    @staticmethod
    def is_config_loaded():
        from airflow.models import Variable
        log.info("Checking if config is already loaded")
        acceldata_lineage_url = Variable.get(ACCELDATA_LINEAGE_URL, None)
        acceldata_lineage_endpoint = Variable.get(ACCELDATA_LINEAGE_ENDPOINT, None)
        acceldata_secret_key = Variable.get(ACCELDATA_SECRET_KEY, None)
        acceldata_access_key = Variable.get(ACCELDATA_ACCESS_KEY, None)
        if acceldata_lineage_url is not None and acceldata_lineage_endpoint is not None and acceldata_secret_key is not None and acceldata_access_key is not None:
            return True
        return False

    @staticmethod
    def find_yaml():
        # Check OPENLINEAGE_CONFIG env variable
        log.info("Trying to set up variables from OPENLINEAGE_CONFIG environment variable.")
        path = os.getenv("OPENLINEAGE_CONFIG", None)
        log.info("OPENLINEAGE_CONFIG path: %s", path)
        try:
            if path and os.path.isfile(path) and os.access(path, os.R_OK):
                log.info("OPENLINEAGE_CONFIG file found.")
                return path
            else:
                log.info("OPENLINEAGE_CONFIG file NOT FOUND.")
        except Exception:
            if path:
                log.exception("Couldn't read file %s: ", path)
            else:
                pass  # We can get different errors depending on system

    @staticmethod
    def persist_url(transport):
        if "url" in transport:
            log.info("URL found in the transport")
            url = transport["url"]
            log.info("Caching ACCELDATA_LINEAGE_URL in Airflow variables")
            Variable.set(ACCELDATA_LINEAGE_URL, url)
            log.info("ACCELDATA_LINEAGE_URL : %s persisted inside Airflow variables.", url)
        else:
            log.error("url not found in OPENLINEAGE_CONFIG")

    @staticmethod
    def persist_endpoint(transport):
        if "endpoint" in transport:
            log.info("endpoint found in the transport")
            endpoint = transport["endpoint"]
            log.info("Caching ACCELDATA_LINEAGE_ENDPOINT in Airflow variables")
            Variable.set(ACCELDATA_LINEAGE_ENDPOINT, endpoint)
            log.info("ACCELDATA_LINEAGE_ENDPOINT : %s persisted inside Airflow variables.", endpoint)
        else:
            log.error("endpoint not found in OPENLINEAGE_CONFIG")

    @staticmethod
    def persist_access_key(auth):
        if "access_key" in auth:
            log.info("endpoint found in the transport")
            access_key = auth["access_key"]
            log.info("Caching ACCELDATA_LINEAGE_URL in Airflow variables")
            Variable.set(ACCELDATA_ACCESS_KEY, access_key)
            log.info("ACCELDATA_ACCESS_KEY : %s persisted inside Airflow variables.", access_key)
        else:
            log.error("access_key not found in OPENLINEAGE_CONFIG")

    @staticmethod
    def persist_secret_key(auth):
        if "secret_key" in auth:
            log.info("endpoint found in the transport")
            secret_key = auth["secret_key"]
            log.info("Caching the OpenLineage ACCELDATA_SECRET_KEY in Airflow variables")
            Variable.set(ACCELDATA_SECRET_KEY, secret_key)
            log.info("ACCELDATA_SECRET_KEY : %s persisted inside Airflow variables.", secret_key)
        else:
            log.error("secret_key not found in OPENLINEAGE_CONFIG")

    def load_config(self):
        log.info("Inside load config. Trying to find the yaml file.")
        file = self.find_yaml()
        if file:
            try:
                log.info("Yaml file detected. Reading config from the openlineage.yml file.")
                with open(file) as f:
                    config: dict[str, Any] = yaml.safe_load(f)
                    log.info("Config data %s: ", config)
                    return config
            except Exception:
                log.warning("Config file not found. Please ensure that config file is setup properly.")
                pass
        return defaultdict(dict)

    @staticmethod
    def get_bearer_token(token):
        return f"Bearer {token}"

    @staticmethod
    def _update_token_to_cache(token: str, expires_in: int):
        expiration = datetime.now() + timedelta(seconds=expires_in)
        log.info("Caching the %s and %s inside airflow variables.", ACCELDATA_BEARER_TOKEN, ACCELDATA_EXPIRES_AT)
        log.info(" %s value is %s", ACCELDATA_EXPIRES_AT, expiration)
        Variable.set(ACCELDATA_BEARER_TOKEN, token)
        Variable.set(ACCELDATA_EXPIRES_AT, expiration)
        log.info("Cached data successfully refreshed.")

    @staticmethod
    def _get_token_from_cache():
        log.info("Fetching the acceldata_bearer_token and its expiration time from Cache")
        cached_token = Variable.get(ACCELDATA_BEARER_TOKEN, None)
        expires_at = Variable.get(ACCELDATA_EXPIRES_AT, None)
        log.info("Cache status has token with %s %s.", ACCELDATA_EXPIRES_AT, expires_at)
        return cached_token, expires_at

    @staticmethod
    def validate_token(cached_token, expires_at):
        # Parse the date string into a datetime object
        log.info("validate_token expires_at: %s", expires_at)
        if expires_at is None:
            log.warning("Token in the cache is absent.")
            return False
        else:
            log.info("Non empty : %s", expires_at)
            expires_at_date = parser.parse(expires_at)
            if cached_token is not None and expires_at_date > datetime.now():
                log.info("Token in the Cache is Valid")
                return True
            else:
                log.warning("Token from the cache has expired.")
                return False

    def _fetch_token_from_admin_central(self) -> tuple[Any, Any] | None:
        headers = {
            'Content-Type': 'application/json',  # Set the content type to JSON
        }

        data = {
            'clientId': CLIENT_ID,
            'secretKey': self.secret_key,
            'accessKey': self.access_key
        }

        token_provider_base_url = Variable.get(ACCELDATA_LINEAGE_URL, None)

        if token_provider_base_url is not None:
            token_provider_url = token_provider_base_url + TOKEN_PROVIDER_PATH
            response = requests.post(token_provider_url, json=data, headers=headers)
            if response.status_code == 200:
                log.info("Fetched the token successfully from Admin Central")
                token_data = response.json()
                token = token_data.get('access_token')
                expires_in = token_data.get('expires_in')
                expires_in_sixty_seconds_ago = expires_in - 60
                return token, expires_in_sixty_seconds_ago
            else:
                log.warning("Failure fetching token from Admin central API. Status code: %s", response.status_code)
                return None, None
        else:
            log.warning("%s not present. Skipping token generation.", ACCELDATA_LINEAGE_URL)

    def refresh_token(self):
        log.info("Refreshing the token data.")
        token, expires_in = self._fetch_token_from_admin_central()
        if token:
            self._update_token_to_cache(token, expires_in)
        else:
            log.error("Token not found in the admin central response.")

    # To persist openlineage.yml configuration to Airflow variables
    def persist_config_to_airflow_variables(self, config):
        log.info("Inside persist_config_to_airflow_variables:")
        from airflow.models import Variable
        if "transport" in config:
            log.info("Transport found in the config")
            transport = config["transport"]
            self.persist_url(transport)
            self.persist_endpoint(transport)

            if "auth" in transport:
                auth = transport["auth"]
                self.persist_access_key(auth)
                self.persist_secret_key(auth)
            else:
                log.error("auth not found in OPENLINEAGE_CONFIG")

        else:
            log.error("transport not found in OPENLINEAGE_CONFIG")

    def get_bearer(self):
        log.info("Checking bearer token in the cache")
        cached_token, expires_at = self._get_token_from_cache()

        is_valid_token = self.validate_token(cached_token, expires_at)
        if is_valid_token:
            return self.get_bearer_token(cached_token)
        else:
            log.info(
                "Either the token doesn't exist or has expired. Refreshing token")
            self.refresh_token()
            cached_token, expires_at = self._get_token_from_cache()
            return self.get_bearer_token(cached_token)



# Copyright 2018-2022 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
import traceback
from typing import List, Optional

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.client.run import Dataset
from openlineage.airflow.utils import get_job_name, try_import_from_string

import logging

_BIGQUERY_NAMESPACE = 'bigquery'

log = logging.getLogger(__name__)


class GCSToBigQueryExtractor(BaseExtractor):
    def __init__(self, operator):
        super().__init__(operator)

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ['GCSToBigQueryOperator']

    def _get_client(self):
        # lazy-load the bigquery Client due to its slow import
        from google.cloud.bigquery import Client
        # Get client using Airflow hook - this way we use the same credentials as Airflow
        log.debug("Get client using Airflow hook - this way we use the same credentials as Airflow")
        if hasattr(self.operator, 'hook') and self.operator.hook:
            log.debug("Airflow Hook found")
            hook = self.operator.hook
            return hook.get_client(
                project_id=hook.project_id,
                location=hook.location
            )
        BigQueryHook = try_import_from_string(
            'airflow.providers.google.cloud.operators.bigquery.BigQueryHook'
        )

        if BigQueryHook is not None:
            log.debug("Biqquery Hook found")
            params = {}
            if hasattr(self.operator, "location"):
                location = self.operator.location
                params["location"] = location
            if hasattr(self.operator, "gcp_conn_id"):
                gcp_conn_id = self.operator.gcp_conn_id
                params["gcp_conn_id"] = gcp_conn_id
            if hasattr(self.operator, "use_legacy_sql"):
                use_legacy_sql = self.operator.use_legacy_sql
                params["use_legacy_sql"] = use_legacy_sql
            if hasattr(self.operator, "impersonation_chain"):
                impersonation_chain = self.operator.impersonation_chain
                params["impersonation_chain"] = impersonation_chain
            if hasattr(self.operator, "delegate_to"):
                params["delegate_to"] = self.operator.delegate_to

            log.debug("params")
            log.debug(params)
            hook = BigQueryHook(**params)
            log.debug("BigQueryHook")
            log.debug(hook)
            return hook.get_client(
                project_id=hook.project_id,
                location=hook.location
            )
        return Client()

    def _get_xcom_bigquery_job_id(self, task_instance):
        bigquery_job_id = task_instance.xcom_pull(
            task_ids=task_instance.task_id, key='job_id')

        self.log.debug(f"bigquery_job_id: {bigquery_job_id}")
        return bigquery_job_id

    def extract(self) -> Optional[TaskMetadata]:
        return None

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
        from openlineage.common.provider.bigquery import BigQueryDatasetsProvider, BigQueryErrorRunFacet

        self.log.debug(f"extract_on_complete({task_instance})")
        self.log.debug(f"extract (GCSToBigQueryOperator)")
        self.log.debug(f"Operator: {self.operator}")

        job_id = None

        try:
            bigquery_job_id = self._get_xcom_bigquery_job_id(task_instance)
            job_id = bigquery_job_id
            if bigquery_job_id is None:
                raise Exception(
                    "Xcom could not resolve BigQuery job id. Job may have failed."
                )
        except Exception as e:
            self.log.error(
                f"Cannot retrieve job details from BigQuery.Client. {e}", exc_info=True
            )
            return TaskMetadata(
                name=f"{self.operator.dag_id}.{self.operator.task_id}",
                run_facets={
                    "bigQuery_error": BigQueryErrorRunFacet(
                        clientError=f"{e}: {traceback.format_exc()}",
                    )
                }
            )

        inputs = []
        outputs = []

        if self.operator.source_objects:
            inputs = [Dataset(
                namespace=f"gs://{self.operator.bucket}",
                name=f"{source_object}",
                facets={},
            ) for source_object in self.operator.source_objects]

        client = self._get_client()

        if job_id is not None:
            job = client.get_job(job_id=job_id)  # type: ignore
            props = job._properties
            log.debug("Job Props")
            log.debug(props)

            if "configuration" in props:
                configuration = props["configuration"]
                if "load" in configuration:
                    load = props["configuration"]["load"]
                    if "destinationTable" in load:
                        destinationTable = load["destinationTable"]
                        if "projectId" in destinationTable:
                            projectId = destinationTable["projectId"]
                            bigquery_destination_project_dataset_table = self.operator.destination_project_dataset_table
                            self.log.debug(
                                f"bigquery destination project dataset table: {bigquery_destination_project_dataset_table}")

                            if bigquery_destination_project_dataset_table:
                                outputs = [Dataset(
                                    namespace=_BIGQUERY_NAMESPACE,
                                    name=f"{projectId}.{bigquery_destination_project_dataset_table}",
                                    facets={},
                                )]

                            taskmetaData = TaskMetadata(
                                name=f"{self.operator.dag_id}.{self.operator.task_id}",
                                inputs=inputs,
                                outputs=outputs
                            )

                            self.log.debug(f"Task Meta Data: {taskmetaData}")
                            return taskmetaData